## 2011-03-24: adapted for RPostgreSQL
## file ok, last checked: 03.05.2007

## checks for differences in the annotation of an updated R annotation package and the annotation in the annotaiton database table.
checkForDifferentAnnotation <- function(con,chip="hgu133plus2",check.columns=c("GenBank","UniGene","LocusLink"),v=TRUE){
  options( useFancyQuotes=FALSE )
  require(pgUtils)
  require("annaffy")
  annotation <- fetch( dbSendQuery(con, paste( "SELECT * FROM annotation WHERE chip=",sQuote(chip) )), n=-1 )
  annotation <- as.matrix(annotation)
  New.annotation <- matrix(ncol=(length(check.columns)+1),nrow=nrow(annotation))
  rownames(New.annotation) <- annotation[,"id"]
  New.annotation[,1] <- annotation[,"id"]
  col.names <- c("id")
  if(v){
    cat("Creating the new annotation table\n")
    progress(0,steps=40)
  }
  for(i in 1:length(check.columns)){
    if(check.columns[i]=="GenBank" | check.columns[i]=="Genbank" | check.columns[i]=="gen_bank"){
      col.names <- c(col.names,"gen_bank")
      New.annotation[,(i+1)] <- getText(aafGenBank(New.annotation[,1],chip))
    }
    if(check.columns[i]=="UniGene" | check.columns[i]=="Unigene" | check.columns[i]=="uni_gene"){
      col.names <- c(col.names,"uni_gene")
      New.annotation[,(i+1)] <- getText(aafUniGene(New.annotation[,1],chip))
    }
    if(check.columns[i]=="LocusLink" | check.columns[i]=="Locuslink" | check.columns[i]=="locuslink"){
      col.names <- c(col.names,"locuslink")
      New.annotation[,(i+1)] <- getText(aafLocusLink(New.annotation[,1],chip))
    }
    if(check.columns[i]=="Symbol" | check.columns[i]=="Name" | check.columns[i]=="symbol"){
      col.names <- c(col.names,"symbol")
      New.annotation[,(i+1)] <- getText(aafSymbol(New.annotation[,1],chip))
    }
    if(v){
      progress(i*100/length(check.columns),steps=40)
    }
  }
  if(v){
    progress(100,steps=40)
    cat(" finished\n\n")
  }
  colnames(New.annotation) <- col.names
  c.length <- ncol(New.annotation)
  diff.table.c <- 2 * c.length
  if(v){
    cat("comparing the annotation\n")
    progress(0,steps=40)
    nrow.ann <- nrow(annotation)
  }
  DiffTable <- NULL
  dummy <- NULL
  for(i in 1:nrow.ann){
    if(min(annotation[i,col.names]==New.annotation[i,col.names])==0){
      if(is.null(DiffTable)){
        DiffTable <- matrix(ncol=diff.table.c,nrow=1)
        DiffTable[1,1:c.length] <- annotation[i,col.names]
        DiffTable[1,(c.length+1):ncol(DiffTable)] <- New.annotation[i,col.names]
      }
      else{
        dummy <- DiffTable
        DiffTable <- matrix(ncol=diff.table.c,nrow=(nrow(dummy)+1))
        DiffTable[1:nrow(dummy),] <- dummy
        DiffTable[nrow(DiffTable),1:c.length] <- annotation[i,col.names]
        DiffTable[nrow(DiffTable),(c.length+1):ncol(DiffTable)] <- New.annotation[i,col.names]
      }
    }
    if(v){
      progress(i*100/nrow.ann,steps=40)
    }
  }
  if(v){
    progress(100,steps=40)
    cat(" finished\n\n")
  }
  colnames(DiffTable) <- c(paste("old",col.names),paste("new",col.names))
  DiffTable
}


dbUpdateAnnotation <- function(Con,data=NULL,chip=NULL,date=NULL,do.backup=TRUE,v=TRUE){
  Suggested = c("id","gen_bank","description","uni_gene","locuslink","symbol","chromosome","chromosome_location","chip","add_info")
  options( useFancyQuotes=FALSE )
  if(is.null(data)){
    stop("Please submit the data that should be inserted into the database as a matrix or data.frame!\n Suggested columns are id, gen_bank, description, uni_gene, locuslink, symbol, chromosome, chromosome_location (which are the annotations provided by Bioconductors annotation packages for Affymetrix GeneChips).")
  }
  if(is.null(chip)){
    stop("Submit a identifier for your chip! For Affymetrix GeneChips for example chip=hgu133plus2.\n\n")
  }
  if(is.null(date)){
    date <- "unknown"
  }
  ##
  ## check if the data has a column named id:
  Colnames.data <- colnames(data)
  if(is.null(Colnames.data)){
    stop("The data frame you submitted does not have column names! Submit a table with column names!")
  }
  if("id" %in% Colnames.data){
  }
  else{
    stop("The submitted data frame must have one column named id containing the unique ids (e.g. Affymetrix pobeset ids).")
  }
  table.list <- dbListTables(Con)
  if( !("annotation" %in% table.list )){
    NewCols <- unique(c(Colnames.data,Suggested))
    if(v){
      cat("creating new annotation table with the following columns:\n")
      for(col in NewCols){
        cat(" ",col,"\n")
      }
    }
    createDBTable( Con, name="annotation", attributes=c(NewCols) )
    Colnames.toinsert <- Colnames.data[Colnames.data %in% NewCols]
  }
  else{
    ## check which columns from the annotation table can be updated with the submitted data.
    Result <- fetch( dbSendQuery(Con,"SELECT * FROM annotation LIMIT 1"), n=-1 )
    Colnames.annotation <- colnames(Result)
    Colnames.toinsert <- Colnames.data[Colnames.data %in% Colnames.annotation]
    if(v){
      NotInsert <- Colnames.data[!(Colnames.data %in% Colnames.annotation)]
      if(length(NotInsert)>0){
        cat("The following columns from the data table are not available in the annotation table in the database. The information from these columns will not be inserted into the table.\n")
        for(col in NotInsert){
          cat(" ",col,"\n")
        }
      }
    }
  }
  if( !("annotation_version" %in% table.list )){
    if(v){
      cat("creating new annotation_version table...\n")
    }
    createDBTable(Con, name="annotation_version", attributes=c("chip","date","update_date","description"), options=c("UNIQUE","","","") )
  }
  new.data <- data[,Colnames.toinsert]
  new.data <- cbind(new.data,chip=chip)
  rownames(new.data) <- new.data[,"id"]
  ##
  ## doing backup of the values in the db...
  if(do.backup){
    table <- fetch( dbSendQuery(Con, paste( "SELECT * FROM annotation_version WHERE chip=",sQuote(chip))  ), n=-1 )
    if(nrow(table)>0){
      if(v){
        cat("Doing backup...\n")
      }
      filename <- chip
      if(nrow(table>0)){
        filename <- paste(chip,table[1,"date"],sep="_")
      }
      filename <- paste(filename,".txt",sep="")
      if(v){
        cat(paste("writing backup to file",filename,"\n"))
      }
      Old.data <- fetch( dbSendQuery(Con,paste( "SELECT * FROM annotation WHERE chip=",sQuote(chip)) ), n=-1 )
      write.table(Old.data,filename,sep="\t",quote=FALSE,row.names=FALSE)
    }
    else{
      if(v){
        cat("Nothing to backup...\n")
      }
    }
  }
  IDs <- NULL
  Test <- fetch( dbSendQuery(Con,paste( "SELECT id FROM annotation WHERE chip=",sQuote(chip)) ), n=-1 )
  if( nrow( Test ) > 0 ){
    IDs <- Test[,"id"]
  }
  BM <- rep(FALSE,nrow(new.data))
  if(length(IDs) > 0){
    if(v){
      cat("updating...")
    }
    BM <- new.data[,"id"] %in% IDs
    updateDBTable( Con, name="annotation", where.col="id", new.values=new.data[BM,], old.values=new.data[BM,"id"])
  } 
  if(max(!BM)==1){
    if(v){
      cat("inserting...")
    }
    insertIntoTable(Con,name="annotation",data=new.data[!BM,])
  }
  if(nrow( dbGetQuery(Con,paste( "SELECT * FROM annotation_version WHERE chip=",sQuote(chip))) )>0){
    updateDBTable( Con, name="annotation_version", where.col="chip", new.values=c(date,date()), old.values=chip, set.col=c("date","update_date"))
  }
  else{
    insertIntoTable( Con, name="annotation_version", attributes=c("chip","date","update_date"), data=c(chip,date,date()))
  }
  if(v){
    cat("finished\n\n")
  }
}


.toMartChip <- function(chip){
	rep <- chip
	if(chip=="hgu133plus2"){
		rep <- "hg_u133_plus_2"
	}
	rep
}


getAnnotation <- function(id=NULL,con=NULL,method="local",chip=NULL,v=TRUE,...){
	match.arg(method,c("local","db","mart"))
	if(method=="local"){
		if(v){
			cat("Using local R annotation packages \n")
			stop("not implemented yet...")
		}
	}
	if(method=="db"){
		if(v){
			cat("Using maDB annotation database \n")
			if(is.null(con)){
				stop("You must connect to the database using the dbConnect function and submit the connection object as con attribute! \n\n")
			}
		}
		result <- dbGetAnnotation(Con=con,id=id,v=v,chip=chip,...)
	}
	if(method=="mart"){
		if(v){
			cat("Using biomaRt annotation \n")
			if(is.null(con)){
				stop("You must connect to mart using the function martConnect and submit the connection object as con attribute! \n\n")
			}
		}
		require(biomaRt)
		martT <- getGene(mart=con,id=id,type=.toMartChip(chip),...)
		Names <- names(martT@table)
		result <- matrix(ncol=(length(Names)+1),nrow=length(martT@id))
		colnames(result) <- c("id",Names)
		rownames(result) <- martT@id
		result[,"id"] <- martT@id
		for(i in 1:length(Names)){
			result[,Names[i]] <- unlist(martT@table[Names[i]])
		}
	}
	result
}


dbGetAnnotation <- function(Con,id,columns=c("gen_bank","description","uni_gene","locuslink","symbol","chromosome"),chip=NULL,search.col="id",v=TRUE,quickload=FALSE){
  require(pgUtils)
  options( useFancyQuotes=FALSE )
  if(is.null(chip)){
    stop("please submit a the chip for which you need the annotation!\n\n")
  }
  if(v){
    cat("Quering the database for the ids...\n")
  }
  result <- fetch( dbSendQuery(Con,paste( "SELECT * FROM annotation_version WHERE chip=",sQuote(chip)) ), n=-1 )
  if(nrow(result)==0){
    stop("No annotation available for this chip!\n")
  }
  if(v){
    cat(paste("The annotation was created on :",result[1,"date"],"\n"))
    cat(paste("The annotation was inserted into the database on ",result[1,"update_date"],"\n"))
  }
  counter <- 0
  if(v){
    cat("Annotating genes\n")
    progress(0,steps=40)
    l.id <- length(id)
  }
  if(length(id)<1000){
    quickload=FALSE
  }
  where <- "id"
  if(search.col=="gen_bank" | search.col=="GenBank" | search.col=="Genbank"){
    where <- "gen_bank"
  }
  if(search.col=="uni_gene" | search.col=="UniGene" | search.col=="Unigene"){
    where <- "uni_gene"
  }
  if(search.col=="LocusLink" | search.col=="locuslink"){
    where <- "locuslink"
  }
  if(search.col=="symbol"){
    where <- "symbol"
  }
  annotation <- NULL
  columns <- c("id",columns)
  if(where=="id" & quickload){
    if(length(id)>100){
      result <- fetch( dbSendQuery(Con, paste( "SELECT * FROM annotation WHERE chip=",sQuote(chip)) ), n=-1 )
    }
    else{
      result <- fetch( dbSendQuery(Con,paste( "SELECT * FROM annotation WHERE chip=",sQuote(chip),"AND id IN (",paste( sQuote(id), collapse=","),")") ), n=-1 )
    }
    if(nrow(result)>0){
      temp.result <- as.matrix(result[,columns])
      rownames(temp.result) <- result[,"id"]
      annotation <- temp.result[id,]
      colnames(annotation) <- columns
    }
  }
  else{
    for(i in 1:length(id)){
      if(v){
        counter <- counter + 1
        progress(counter*100/l.id,steps=40)
      }
      result <- fetch( dbSendQuery(Con,paste( "SELECT * FROM annotation WHERE chip=",sQuote(chip),"AND",where,"=" ,sQuote(id[i])) ), n=-1 )
      if(nrow(result)>0){
        temp.result <- as.matrix(result[,columns])
        if(is.null(annotation)){
          annotation <- temp.result
        }
        else{
          new.annotation <- matrix(ncol=length(columns),nrow=(length(annotation[,1])+nrow(result)))
          new.annotation[1:length(annotation[,1]),] <- annotation
          new.annotation[(length(annotation[,1])+1):length(new.annotation[,1]),] <- temp.result
          colnames(new.annotation) <- columns
          annotation <- new.annotation
        }
      }
    }
  }
  if(v){
    progress(100,steps=40)
    cat("\nfinished.\n")
  }
  annotation
}



